'''Winny Commands.

See also Isamu Kaneko: The Technology of Winny.
http:/www.amazon.co.jp/exec/obidos/ASIN/4756145485

Command code table (copied from Poeny):
    PROTOCOL_HEADER                 =  0,
    SPEED                           =  1,
    CONNECTION_TYPE                 =  2,
    NODE_DETAILS                    =  3,
    ANOTHER_NODE                    =  4,
    BBS_PORT                        =  5,
    DIFFUSION_REQUEST               = 10,
    FILE_REQUEST                    = 11,
    CONDITIONAL_DIFFUSION_REQUEST   = 12,
    QUERY                           = 13,
    FILE_RESPONSE                   = 21,
    CLOSE                           = 31,
    CONNECTED_LIMITATION            = 32,
    WRONG_LISTENING_PORT            = 33,
    REJECT                          = 34,
    SLOW_RATE                       = 35,
    LIAR                            = 36,
    LOW_VERSION                     = 97,
'''
#
# Copyright (c) 2006 Pyny Project.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: $
#

import struct

import rc4
from nyexcept import *

__version__ = '$Revision: $'
__all__ = ['NyProtocolHeader',
           'NySpeed']

int_size = 4                # byte
float_size = 4 
command_length_size = 4     # bytes
code_size = 1               # bytes
header_length = command_length_size + code_size
major_version = 'Winny Ver2.0b1 (poeny)'
minor_version = 12710

def make_step2key(step1key):
    '''What is this?
    '''
    step2key = []
    for c in step1key:
        step2key.append(chr(ord(c)^0x39))
    return ''.join(step2key)

def get_cstring(s):
    '''Parse string and get C string.

    C string ends with 0x00.
    '''
    length = s.find(chr(0))
    if length >= 0:
        return s[:length]
    else:
        return s

def int_to_packet(n):
    '''Make int packet form.

    Winny packet is little endian.

    sample:
    >>> from conv import hexstr
    >>> hexstr(int_to_packet(123456789))
    '15cd5b07'
    '''
    array = ['x', 'x', 'x', 'x']
    array[3] = chr((n & 0xff000000) >> 24)
    array[2] = chr((n & 0x00ff0000) >> 16)
    array[1] = chr((n & 0x0000ff00) >> 8)
    array[0] = chr((n & 0x000000ff))
    return ''.join(array)


class NyCommand:
    '''Winny Command.
    '''
    length = 0
    code = -1
    data = None
    header_length = 0

    def __len__(self):
        return length

    def unpack(self, command_block, unpack_all=True):
        '''Unpack packet style command.

        Packet is string (byte array).
        If unpack_all is False, unpack header only.
        '''
        if len(command_block) < header_length:
            raise CommandError('NyCommand: command is too small')
        self.length = len(command_block)
        if self.length == 0:
            raise CommandError('NyCommand: command code is nothing')
        self.gotcode = command_block[command_length_size]
        if len(command_block) < header_length + self.length - 1:
            #raise CommandError('NyCommand: length is too small')
            #XXX
            pass
        if not unpack_all:
            self.data = None
            if self.length == 1:
                return None
            else:
                return command_block[header_length:]
        if self.length > header_size:
            self.data = command_block[header_length:]
        else:
            self.data = None
        return self.data

    def assign(self, command):
        self.length = command.length
        self.code = command.code
        self.data = command.data

    def pack(self):
        '''Make packet, byte stream.
        '''
        packet = []
        packet.extend(int_to_packet(code_size + len(self.data)))
        packet.append(chr(self.code))
        packet.extend(self.data)
        self.data = None
        return ''.join(packet)

    def next_index(self):
        '''Next command index of packet steram.
        '''
        return header_size + self.length - 1

    def __str__(self):
        return 'CommandCode(%d) Length(%d)' % (self.code, self.length)

# End of NyCommand


class NyRawCommand(NyCommand):
    def __init__(self, command_block):
        self.unpack(command_block)

    def unpack(self, command):
        if isinstance(command, NyRawCommand):
            self.assign(command)
        else:
            self.data = NyCommand.unpack(self, command)

# End of NyRawCommand


class UnpackMixIn:
    '''Unpack common method.
    '''
    def __init__(self, raw_command=None):
        if raw_command is not None:
            self.unpack(raw_command)

    def unpack(self, command):
        if isinstance(command, NyRawCommand):
            self.assign(command)
            self.unpack_internal(command.data)
        else:
            data = NyCommand.unpack(self, command, False)
            self.unpack_internal(data)

# End of UnpackMixIn


class NyProtocolHeader(UnpackMixIn, NyCommand):
    '''Command00 - Winny Protocol Header.

    Records: <00><version><password>.

    Sample:
    >>> from conv import hexstr
    >>> h = NyProtocolHeader()
    >>> h.major, h.minor = 'Winny Ver2.0b1 (poeny)', 12710
    >>> str(h)
    'Winny Ver2.0b1 (poeny)(12710)'
    >>> data = h.pack()
    >>> hexstr(data)
    '1b000000005f9c51446217f43658711c5eb7332084e7bb9b559276f3d9c77f'
    >>> h = NyProtocolHeader(data)
    >>> str(h)
    'Winny Ver2.0b1 (poeny)(12710)'
    '''
    code = 0
    cert_key = '\x39\x38\x37\x38\x39\x61\x73\x6a'
    major = ''
    minor = 0

    def cert_crypt(self, data):
        return rc4.crypt(self.cert_key, data)

    def unpack_internal(self, data):
        if self.length < int_size:
            raise CommandError('Command is too small')
        cert = self.cert_crypt(data)
        # minor is 32bit little endian
        self.minor = struct.unpack('l', cert[:int_size])[0]
        self.major = cert[int_size:]
        self.data = None

    def __str__(self):
        return '%s(%d)' % (self.major, self.minor)

    def pack(self):
        data = int_to_packet(self.minor) + self.major
        self.data = self.cert_crypt(data)
        return NyCommand.pack(self)

# End of NyProtocolHeader


class NySpeed(UnpackMixIn, NyCommand):
    '''Command01 - Report Line Speed.

    Records: <01><speed (KB/s)>.

    Sample:
    >>> from conv import hexstr
    >>> com = NySpeed()
    >>> com.speed = 120
    >>> data = com.pack()
    >>> hexstr(data)
    '05000000010000f042'
    >>> com = NySpeed(data)
    >>> com.speed
    120.0
    '''
    code = 1
    speed = 0

    def unpack_internal(self, data):
        if len(data) < float_size:
            raise CommandError('Speed: length is too small')
        self.speed = struct.unpack('f', data)[0]
        self.data = None

    def pack(self):
        self.data = struct.pack('f', self.speed)
        return NyCommand.pack(self)

# End of NySpeed


class NyConnectionType(UnpackMixIn, NyCommand):
    '''Command02 - Connection Type.

    Recods: <02><port0 flag><badport0 flag><bbslink flag>.

    Sample:
    >>> from conv import hexstr
    >>> com = NyConnectionType()
    >>> com.setvalues(linktypestr='Transfer')
    >>> com.isport0, com.isbadport0, com.isbbslink = True, False, True
    >>> data = com.pack()
    >>> hexstr(data)
    '050000000201010001'
    >>> com = NyConnectionType(data)
    >>> com.linktypestr, com.isport0, com.isbadport0, com.isbbslink
    ('Transfer', True, False, True)
    '''
    code = 2
    linktypes = {'Search': 0, 'Transfer': 1, 'BbsSearch': 2}
    linktype = -1
    linktypestr = ''
    isport0 = False
    isbadport0 = False
    isbbslink = False

    def unpack_internal(self, data):
        if len(data) < 4:  #XXX 4 flags
            raise CommandError('ConnectionType: length is too small')
        linktype = ord(data[0])
        for k in self.linktypes:
            if self.linktypes[k] == linktype:
                self.linktypestr = k
                break
        else:
            raise CommandError(
                    'ConnectionType: unknown LinkType "%d"' % linktype)
        self.isport0 = (data[1] != chr(0))
        self.isbadport0 = (data[2] != chr(0))
        self.isbbslink = (data[3] != chr(0))
        self.data = None

    def setvalues(self, linktypestr=''):
        if linktypestr:
            self.linktypestr = linktypestr
            self.linktype = self.linktypes[linktypestr]

    def pack(self):
        self.data = [chr(self.linktype),
                     chr(int(self.isport0)),
                     chr(int(self.isbadport0)),
                     chr(int(self.isbbslink))]
        return NyCommand.pack(self)

# End of NyConnectionType


def _test():
    import doctest, nycommand
    return doctest.testmod(nycommand)
    pass

if __name__ == '__main__':
    _test()
